import pkg_resources

from tmdbapis.tmdb import TMDbAPIs
from tmdbapis.objs.simple import AlternativeName, AlternativeTitle, Certification, Country, CountryCertifications, \
    CountryWatchProviders, Department, FindResults, Genre, Group, Language, ReleaseDate, Timezones, Trailer, \
    Translation, User, Video, WatchProvider
from tmdbapis.objs.image import TMDbImage, Backdrop, Logo, Poster, Profile, Still, Tagged
from tmdbapis.objs.reload import TMDbReload, Account, Collection, Company, Configuration, Credit, Episode, \
    EpisodeGroup, Keyword, Movie, Network, Person, Review, Season, TVShow
from tmdbapis.exceptions import TMDbException, Invalid, NotFound, Unauthorized, Authentication

try:
    __version__ = pkg_resources.get_distribution("tmdbapis").version
except pkg_resources.DistributionNotFound:
    __version__ = ""
__author__ = "Nathan Taggart"
__credits__ = "meisnate12"
__package_name__ = "tmdbapis"
__project_name__ = "TMDbAPIs"
__description__ = "A lightweight Python library for The V3 and V4 TMDb APIs."
__url__ = "https://github.com/Kometa-Team/TMDbAPIs"
__email__ = "kometateam@proton.me"
__license__ = "MIT License"
__all__ = [
    "TMDbAPIs",
    "AlternativeName",
    "AlternativeTitle",
    "Certification",
    "Country",
    "CountryCertifications",
    "CountryWatchProviders",
    "Department",
    "FindResults",
    "Genre",
    "Group",
    "Language",
    "ReleaseDate",
    "Timezones",
    "Trailer",
    "Translation",
    "User",
    "Video",
    "WatchProvider",
    "TMDbImage",
    "Backdrop",
    "Logo",
    "Poster",
    "Profile",
    "Still",
    "Tagged",
    "TMDbReload",
    "Account",
    "Collection",
    "Company",
    "Configuration",
    "Credit",
    "Episode",
    "EpisodeGroup",
    "Keyword",
    "Movie",
    "Network",
    "Person",
    "Review",
    "Season",
    "TVShow",
    "TMDbException",
    "Invalid",
    "NotFound",
    "Unauthorized",
    "Authentication"
]



from abc import ABC, abstractmethod
from typing import Any, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from tmdbapis.tmdb import TMDbAPIs
    from tmdbapis.api3 import API3
    from tmdbapis.api4 import API4


class TMDbObj(ABC):
    """ Base Class for TMDb Objects. """

    def __init__(self, tmdb: "TMDbAPIs", data):
        self._loading = True
        self._partial = False
        self._name = None
        self._tmdb = tmdb
        self._api: "API3" = tmdb._api if tmdb else None
        self._api4: "API4" = tmdb._api4 if tmdb else None
        self._load(data)

    @abstractmethod
    def _load(self, data):
        self._data = data
        self._loading = True
        self.id = None

    def _finish(self, name):
        self._name = str(name)
        self._loading = False

    def _image_url(self, image_path):
        return f"{self._tmdb._image_url}{image_path}" if image_path else None

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.id:
            return f"[{self.id}:{self._name}]"
        else:
            return self._name

    def __eq__(self, other):
        if type(self) is type(other):
            if self.id is None and other.id is None:
                return self._name == other._name
            elif self.id is not None and other.id is not None:
                return self.id == other.id
            else:
                return False
        elif isinstance(other, int) and self.id is not None:
            return self.id == other
        else:
            return str(self._name) == str(other)

    def __getattribute__(self, item):
        value = super().__getattribute__(item)
        if item.startswith("_") or self._loading or not self._partial or \
                (isinstance(value, (list, dict, int, float)) and value) or \
                (not isinstance(value, (list, dict, int, float)) and value is not None):
            return value
        self._load(None)
        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if key.startswith("_") or self._loading:
            super().__setattr__(key, value)
        else:
            raise AttributeError("Attributes cannot be edited")

    def __delattr__(self, key):
        raise AttributeError("Attributes cannot be deleted")

    def _parse(self, data=None, attrs: Optional[Union[str, list]] = None, value_type: str = "str",
               default_is_none: bool = False, is_list: bool = False, is_dict: bool = False, extend: bool = False,
               key: Any = None):
        """ Validate the value given from the options given.
    
            Parameters:
                attrs (Optional[Union[str, list]]): check data for these attributes.
                value_type (str): Type that the value is.
                default_is_none (bool): Makes default None.
                is_list (bool): value is list of values.
                is_dict (bool): value is dict of values.
                extend (bool): value is list of values.
                key (Any): extra key.

            Returns:
                Any: Parsed Value
        """
        return self._tmdb._parse(data=self._data if data is None else data, attrs=attrs, value_type=value_type,
                                 default_is_none=default_is_none, is_list=is_list, is_dict=is_dict, extend=extend, key=key)


import os, sys, time, unittest
from datetime import datetime, timedelta
from dotenv import load_dotenv
from github import Auth, Github
from tmdbapis.api3 import API3
from tmdbapis.exceptions import NotFound, Invalid, Authentication
from tmdbapis.objs.reload import Movie, Person, TVShow
from tmdbapis.tmdb import TMDbAPIs

"""
import logging
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())
"""

load_dotenv()

apikey = os.environ["TMDB_APIKEY"]
session_id = os.environ["TMDB_SESSION"]
v4 = os.environ["TMDB_V4_TOKEN"]
access = os.environ["TMDB_V4_ACCESS"]
username = os.environ["TMDB_USERNAME"]
password = os.environ["TMDB_PASSWORD"]
gh_token = os.environ["PAT"]
local = os.environ["LOCAL"] == "True"
py_version = f"{sys.version_info.major}.{sys.version_info.minor}"

movie_ids = {
    "3.8": {"id": 1892, "name": "Return of the Jedi"},
    "3.9": {"id": 1893, "name": "Star Wars: Episode I - The Phantom Menace"},
    "3.10": {"id": 1894, "name": "Star Wars: Episode II - Attack of the Clones"},
    "3.11": {"id": 1895, "name": "Star Wars: Episode III - Revenge of the Sith"},
    "3.12": {"id": 140607, "name": "Star Wars: The Force Awakens"},
}

show_ids = {
    "3.8": {"id": 4194, "name": "Star Wars: The Clone Wars"},
    "3.9": {"id": 105971, "name": "Star Wars: The Bad Batch"},
    "3.10": {"id": 60554, "name": "Star Wars Rebels"},
    "3.11": {"id": 203085, "name": "Star Wars: Tales of the Jedi"},
    "3.12": {"id": 83867, "name": "Star Wars: Andor"},
}

episode_ids = {
    "3.8": {"id": 1, "name": "Ambush"},
    "3.9": {"id": 2, "name": "Rising Malevolence"},
    "3.10": {"id": 3, "name": "Shadow of Malevolence"},
    "3.11": {"id": 4, "name": "Destroy Malevolence"},
    "3.12": {"id": 5, "name": "Rookies"},
}

class APITests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.api = TMDbAPIs(apikey, v4_access_token=access, session_id=session_id)
        cls.api_v4 = TMDbAPIs(apikey, v4_access_token=access)
        cls.api_v3_session = TMDbAPIs(apikey)
        cls.api_v4_session = TMDbAPIs(apikey)
        cls.raw = API3(apikey, session_id=session_id)

    def test_aa_session(self):
        self.assertIsNotNone(self.api_v4.session_id)

        self.api_v3_session.language = "en"
        self.assertEqual(self.api_v3_session.language, "en")
        self.api_v3_session.language = "en-US"
        self.assertEqual(self.api_v3_session.language, "en-US")
        self.api_v3_session.language = None
        self.assertIsNone(self.api_v3_session.language)
        with self.assertRaises(Invalid):
            self.api_v3_session.language = "invalid_value"

        self.api_v3_session.include_language = "en,null,en-US"
        self.assertEqual(self.api_v3_session.include_language, "en,null,en-US")
        self.api_v3_session.include_language = None
        self.assertIsNone(self.api_v3_session.include_language)
        with self.assertRaises(Invalid):
            self.api_v3_session.include_language = "invalid_value"

        with self.assertRaises(Authentication):
            print(self.api_v3_session.account_id)
        self.api_v3_session.authenticate(username, password)
        self.api_v4_session._api._session_id = self.api_v3_session.session_id

        with self.assertRaises(Authentication):
            self.api_v4_session.v4_approved()
        account = self.api_v4_session.account()
        with self.assertRaises(Authentication):
            account.movie_recommendations()
        self.api_v4_session.v4_access(v4)
        with self.assertRaises(Authentication):
            account.movie_recommendations()
        v4_url = self.api_v4_session.v4_authenticate()
        if local:
            print(f"\n\nApprove URL: {v4_url}")
            with open("url.txt", "w") as file1:
                file1.write(v4_url)
            time.sleep(120)
        else:
            git = Github(auth=Auth.Token(gh_token))
            repo = git.get_user("Kometa-Team").get_repo("TMDbAPIs")
            issue_id = repo.create_issue(f"Run: {datetime.now().strftime('%Y-%m-%d %H:%M')} {py_version}", body=v4_url,
                                         assignee="meisnate12", labels=["approval-needed"]).number
            while True:
                issue = repo.get_issue(issue_id)
                if "approval-needed" not in [la.name for la in issue.labels]:
                    issue.edit("Delete", body="delete", state="closed")
                    break
                time.sleep(10)
        self.api_v4_session.v4_approved()
        print("\ntest_aa_session: ", end="")

    def test_ab_variables(self):
        print("\ntest_ab_variables: ", end="")
        self.assertEqual(self.api.session_id, session_id)
        self.assertIsNotNone(self.api.v4_access_token)
        self.assertEqual(self.api.account_id, 8568268)
        self.assertEqual(self.api.v4_account_id, "5d339fb42f8d097bccd118c7")

    def test_account(self):
        print("\ntest_account: ", end="")
        account = self.api.account()
        self.assertEqual(account.id, 8568268)
        self.assertEqual(account.username, "meisnate12")
        self.assertGreater(len(account.created_lists(v3=True).results), 0)
        self.assertGreater(len(account.created_lists().results), 0)
        self.assertGreater(len(account.favorite_movies(v3=True).results), 0)
        self.assertGreater(len(account.favorite_movies().results), 0)
        self.assertGreater(len(account.favorite_tv_shows(v3=True).results), 0)
        self.assertGreater(len(account.favorite_tv_shows().results), 0)
        self.assertGreater(len(account.rated_movies(v3=True).results), 0)
        self.assertGreater(len(account.rated_movies().results), 0)
        self.assertGreater(len(account.rated_tv_shows(v3=True).results), 0)
        self.assertGreater(len(account.rated_tv_shows().results), 0)
        self.assertGreater(len(account.rated_episodes().results), 0)
        self.assertGreater(len(account.movie_watchlist(v3=True).results), 0)
        self.assertGreater(len(account.movie_watchlist().results), 0)
        self.assertGreater(len(account.tv_show_watchlist(v3=True).results), 0)
        self.assertGreater(len(account.tv_show_watchlist().results), 0)
        self.assertGreater(len(account.movie_recommendations().results), 0)
        self.assertGreater(len(account.tv_show_recommendations().results), 0)

    def test_certifications(self):
        print("\ntest_certifications: ", end="")
        movie = self.api.movie_certifications()
        tv = self.api.tv_certifications()
        self.assertIn("US", movie)
        self.assertIn("US", tv)
        self.assertIn("Certifications", f"{movie['US']}")

    def test_changes(self):
        print("\ntest_changes: ", end="")
        now = datetime.now()
        with self.assertRaises(Invalid):
            self.api.movie_change_list(start_date=f"fa/01/{now.year}")
        self.assertGreater(len(self.api.movie_change_list(start_date=now - timedelta(days=7), end_date=now)), 0)
        self.assertGreater(len(self.api.movie_change_list()), 0)
        self.assertGreater(len(self.api.tv_change_list()), 0)
        self.assertGreater(len(self.api.person_change_list()), 0)

    def test_collections(self):
        print("\ntest_collections: ", end="")
        collection = self.api.collection(10)
        with self.assertRaises(AttributeError):
            collection.name = "New Name"
        with self.assertRaises(AttributeError):
            del collection.name
        self.assertTrue(collection == collection)
        self.assertTrue(collection == 10)
        self.assertTrue(collection == "Star Wars Collection")
        self.assertEqual(collection.name, "Star Wars Collection")
        self.assertGreater(len(collection.movies), 0)

    def test_companies(self):
        print("\ntest_companies: ", end="")
        company = self.api.company(1)
        self.assertEqual(company.name, "Lucasfilm Ltd.")
        self.assertGreater(len(company.movies), 0)
        self.assertGreater(len(company.tv_shows), 0)

    def test_credit(self):
        print("\ntest_credit: ", end="")
        credit = self.api.credit("52fe420dc3a36847f8000441")
        self.assertEqual(credit.name, "Mark Hamill")
        self.assertEqual(credit.person_id, 2)
        self.assertEqual(credit.character, "Luke Skywalker")
        credit2 = self.api.credit("5beb247f92514143e6058194")
        self.assertEqual(credit2.name, "Pedro Pascal")

    def test_discover(self):
        print("\ntest_discover: ", end="")
        with self.assertRaises(Invalid):
            self.api.discover_movies(invalid_attr=1)
        with self.assertRaises(Invalid):
            self.api.discover_movies(sort_by="invalid_value")
        with self.assertRaises(Invalid):
            self.api.discover_movies(region="invalid_value")
        with self.assertRaises(Invalid):
            self.api.discover_movies(include_video="y", certification_country="invalid_value")
        with self.assertRaises(Invalid):
            self.api.discover_movies(include_video="n", certification_country="invalid_value")
        with self.assertRaises(Invalid):
            self.api.discover_movies(certification="invalid_value")
        with self.assertRaises(Invalid):
            self.api.discover_movies(**{"release_date.gte": datetime.now(), "release_date.lte": "gg/gg/gggg"})
        with self.assertRaises(Invalid):
            self.api.discover_movies(**{"vote_average.gte": 5, "vote_average.lte": 0})
        with self.assertRaises(Invalid):
            self.api.discover_movies(**{"vote_count.gte": 5, "vote_count.lte": 0})
        with self.assertRaises(Invalid):
            self.api.discover_movies(certification="PG", certification_country="US", with_watch_monetization_types="invalid_value")
        with self.assertRaises(Invalid):
            self.api.discover_movies(with_watch_monetization_types="free")
        with self.assertRaises(Invalid):
            self.api.discover_movies(with_watch_monetization_types="free", watch_region="US", include_video="g")
        movies = self.api.discover_movies(with_companies=1, sort_by="popularity.asc")
        self.assertEqual(len(movies.results), 20)
        movies.load_next()
        self.assertEqual(len(movies.results), 20)
        with self.assertRaises(Invalid):
            movies.load_page(0)
        with self.assertRaises(Invalid):
            movies.get_results(-1)
        self.assertEqual(len(movies.get_results(50)), 50)
        self.assertGreater(len(movies.get_results()), 50)
        movies.load_page(1)
        self.assertIsNotNone([m for m in movies])
        self.assertIsNotNone(movies[1])
        tv_shows = self.api.discover_tv_shows(with_companies=1)
        self.assertGreater(len(tv_shows), 0)

    def test_find(self):
        print("\ntest_find: ", end="")
        self.assertGreater(len(self.api.find_by_id(imdb_id="tt0076759").movie_results), 0)
        self.assertGreater(len(self.api.find_by_id(facebook_id="StarWars").movie_results), 0)
        self.assertGreater(len(self.api.find_by_id(twitter_id="starwars").movie_results), 0)
        self.assertGreater(len(self.api.find_by_id(instagram_id="starwars").movie_results), 0)
        self.assertGreater(len(self.api.find_by_id(freebase_mid="/m/0524b41").tv_results), 0)
        self.assertGreater(len(self.api.find_by_id(freebase_id="/en/game_of_thrones").tv_results), 0)
        self.assertGreater(len(self.api.find_by_id(tvdb_id="121361").tv_results), 0)
        self.assertGreater(len(self.api.find_by_id(tvrage_id="24493").tv_results), 0)
        with self.assertRaises(Invalid):
            self.api.find_by_id()
        with self.assertRaises(NotFound):
            self.api.find_by_id(imdb_id="tt00764596759")

    def test_genres(self):
        print("\ntest_genres: ", end="")
        self.assertGreater(len(self.api.movie_genres(reload=True)), 0)
        self.assertGreater(len(self.api.tv_genres(reload=True)), 0)

    def test_keyword(self):
        print("\ntest_keyword: ", end="")
        keyword = self.api.keyword(4270)
        self.assertEqual(keyword.name, "galaxy")
        self.assertGreater(len(keyword.movies), 0)
        self.assertGreater(len(keyword.tv_shows), 0)

    def test_lists(self):
        print("\ntest_lists: ", end="")
        with self.assertRaises(Invalid):
            self.api.create_list("PMM Fail", "fail", load=True)
        for test_list in [
            self.api_v3_session.create_list("PMM Test V3", "en", load=True),
            self.api.create_list("PMM Test V4", self.api._iso_639_1["en"], load=True)
        ]:
            test_list.add_items([(524434, "movie")])
            self.assertGreater(len(test_list), 0)
            self.assertTrue(test_list.has_item((524434, "movie")))
            test_list.remove_items((524434, "movie")) # noqa
            self.assertFalse(test_list.has_item((524434, "movie")))
            test_list.add_items((524434, "movie")) # noqa
            if "V4" in test_list.name:
                with self.assertRaises(Invalid):
                    test_list.update()
                with self.assertRaises(Invalid):
                    test_list.update(sort_by="not valid")
                test_list.update(name="PMM Test V4 V2", description="Test Description")
                self.assertEqual(test_list.name, "PMM Test V4 V2")
                self.assertEqual(test_list.description, "Test Description")
                test_list.update_items(((524434, "movie"), "test description")) # noqa

            test_list.clear()
            self.assertFalse(test_list.has_item((524434, "movie")))
            self.assertEqual(len(test_list), 0)
            test_list.delete()
            with self.assertRaises(NotFound):
                test_list.reload()

    def test_movie(self):
        print("\ntest_movie: ", end="")
        with self.assertRaises(NotFound):
            self.api.movie(10)
        with self.assertRaises(Invalid):
            self.api.now_playing_movies(region="fail")
        movie = self.api.movie(movie_ids[py_version]["id"])
        self.assertIsNotNone(str(movie.watch_providers))
        self.assertEqual(movie.title, movie_ids[py_version]["name"])
        self.assertIsInstance(self.api.latest_movie(), Movie)
        self.assertGreater(len(self.api.now_playing_movies(region="us")), 0)
        self.assertGreater(len(self.api.popular_movies(region=self.api._iso_3166_1["us"])), 0)
        self.assertGreater(len(self.api.top_rated_movies()), 0)
        self.assertGreater(len(self.api.upcoming_movies()), 0)
        with self.assertRaises(Invalid):
            movie.rate(11.0)
        movie.rate(8.5)
        time.sleep(2)
        movie.mark_as_favorite()
        time.sleep(2)
        movie.add_to_watchlist()
        time.sleep(2)
        movie.reload()
        self.assertEqual(movie.rated, 8.5)
        self.assertTrue(movie.favorite)
        self.assertTrue(movie.watchlist)
        time.sleep(2)
        movie.delete_rating()
        time.sleep(2)
        movie.remove_as_favorite()
        time.sleep(2)
        movie.remove_from_watchlist()
        time.sleep(2)
        movie.reload()
        self.assertIsNone(movie.rated)
        self.assertFalse(movie.favorite)
        self.assertFalse(movie.watchlist)
        movie.lists.load_page(1)
        self.assertGreater(len(movie.lists), 0)
        movie.reviews.load_page(1)
        self.assertGreater(len(movie.reviews), 0)
        movie.recommendations.load_next()
        self.assertGreater(len(movie.recommendations), 0)
        movie.similar.load_next()
        self.assertGreater(len(movie.similar), 0)

    def test_movie_lists(self):
        print("\ntest_movie_lists: ", end="")
        with self.assertRaises(Invalid):
            self.api_v3_session.movie_watchlist(sort_by="invalid_value")
        with self.assertRaises(Invalid):
            self.api.movie_recommendations(sort_by="invalid_value")
        with self.assertRaises(Invalid):
            self.api.tv_show_recommendations(sort_by="invalid_value")
        self.assertGreater(len(self.api.movie_recommendations()), 0)

    def test_network(self):
        print("\ntest_network: ", end="")
        network = self.api.network(1)
        self.assertEqual(network.name, "Fuji TV")
        self.assertGreater(len(network.tv_shows), 0)

    def test_person(self):
        print("\ntest_person: ", end="")
        person = self.api.person(2)
        self.assertEqual(person.name, "Mark Hamill")
        self.assertIn("Tagged", str(person.tagged))
        self.assertIsInstance(self.api.latest_person(), Person)
        self.assertGreater(len(self.api.popular_people()), 0)

    def test_providers(self):
        print("\ntest_providers: ", end="")
        self.assertGreater(len(self.api.provider_regions()), 0)
        movie_providers = self.api.movie_providers()
        self.assertIn(movie_providers[0].name, f"{movie_providers[0]}")
        self.assertGreater(len(movie_providers), 0)
        self.assertGreater(len(self.api.tv_providers()), 0)

    def test_review(self):
        print("\ntest_review: ", end="")
        review = self.api.review("58a231c5925141179e000674")
        self.assertEqual(review.author, "Cat Ellington")

    def test_search(self):
        print("\ntest_search: ", end="")
        movies = self.api.movie_search("The Lord of the Rings")
        self.assertGreater(movies.total_results, 0)
        self.assertEqual(len(movies.results), 20)
        movies.load_next()
        self.assertGreater(len(movies.results), 0)
        with self.assertRaises(NotFound):
            movies.load_next()
        with self.assertRaises(NotFound):
            self.api.movie_search("fgdhsdfghsdfghsdgf")

        self.assertGreater(self.api.company_search("Disney").total_results, 0)
        with self.assertRaises(NotFound):
            self.api.company_search("fgdhsdfghsdfghsdgf")
        self.assertGreater(self.api.collection_search("Star Wars").total_results, 0)
        with self.assertRaises(NotFound):
            self.api.collection_search("fgdhsdfghsdfghsdgf")
        self.assertGreater(self.api.keyword_search("fantasy").total_results, 0)
        with self.assertRaises(NotFound):
            self.api.keyword_search("fgdhsdfghsdfghsdgf")
        self.assertGreater(self.api.multi_search("Star Wars").total_results, 0)
        with self.assertRaises(NotFound):
            self.api.multi_search("fgdhsdfghsdfghsdgf")
        self.assertGreater(self.api.people_search("George Lucas").total_results, 0)
        with self.assertRaises(NotFound):
            self.api.people_search("fgdhsdfghsdfghsdgf")
        self.assertGreater(self.api.tv_search("Star Wars").total_results, 0)
        with self.assertRaises(NotFound):
            self.api.tv_search("fgdhsdfghsdfghsdgf")

    def test_trending(self):
        print("\ntest_trending: ", end="")
        self.assertGreater(len(self.api.trending("movie", "day")), 0)
        self.assertGreater(len(self.api.trending("tv", "day")), 0)
        self.assertGreater(len(self.api.trending("movie", "week")), 0)
        self.assertGreater(len(self.api.trending("tv", "week")), 0)
        with self.assertRaises(Invalid):
            self.api.trending("test", "day")
        with self.assertRaises(Invalid):
            self.api.trending("all", "test")

    def test_tv_episode(self):
        print("\ntest_tv_episode: ", end="")
        episode = self.api.tv_episode(4194, 1, episode_ids[py_version]["id"])
        self.assertEqual(episode.name, episode_ids[py_version]["name"])
        with self.assertRaises(Invalid):
            episode.rate(11.0)
        episode.rate(8.5)
        time.sleep(2)
        episode.reload()
        time.sleep(2)
        self.assertEqual(episode.rated, 8.5)
        time.sleep(4)
        episode.delete_rating()
        time.sleep(4)
        episode.reload()
        self.assertIsNone(episode.rated)

    def test_tv_episode_group(self):
        print("\ntest_tv_episode_group: ", end="")
        episode_group = self.api.episode_group("5b11ba820e0a265847002c6e")
        self.assertEqual(episode_group.name, "Chronological Order")

    def test_tv_season(self):
        print("\ntest_tv_season: ", end="")
        season = self.api.tv_season(4194, 1)
        self.assertEqual(season.name, "Season 1")

    def test_tv_show(self):
        print("\ntest_tv_show: ", end="")
        show = self.api.tv_show(show_ids[py_version]["id"])
        self.assertEqual(show.name, show_ids[py_version]["name"])
        self.assertIsInstance(self.api.latest_tv(), TVShow)
        self.assertGreater(len(self.api.tv_airing_today()), 0)
        self.assertGreater(len(self.api.tv_on_the_air()), 0)
        self.assertGreater(len(self.api.popular_tv()), 0)
        self.assertGreater(len(self.api.top_rated_tv()), 0)
        with self.assertRaises(Invalid):
            show.rate(11.0)
        show.rate(8.5)
        time.sleep(2)
        show.mark_as_favorite()
        time.sleep(2)
        show.add_to_watchlist()
        time.sleep(2)
        show.reload()
        self.assertEqual(show.rated, 8.5)
        self.assertTrue(show.favorite)
        self.assertTrue(show.watchlist)
        time.sleep(2)
        show.delete_rating()
        time.sleep(2)
        show.remove_as_favorite()
        time.sleep(2)
        show.remove_from_watchlist()
        time.sleep(2)
        show.reload()
        self.assertIsNone(show.rated)
        self.assertFalse(show.favorite)
        self.assertFalse(show.watchlist)
        show.recommendations.load_next()
        self.assertGreater(len(show.recommendations), 0)
        show.similar.load_next()
        self.assertGreater(len(show.similar), 0)

    def test_zz_logout(self):
        print("\ntest_logout: ", end="")
        self.api_v4_session.logout()



from abc import abstractmethod
from typing import Optional, Union, List, Tuple
from urllib.parse import urlparse, parse_qs

from tmdbapis.objs.base import TMDbObj
from tmdbapis.objs.reload import Movie, TVShow
from tmdbapis.exceptions import NotFound, Invalid, TMDbException

v3_sorts = ["created_at.asc", "created_at.desc"]
v4_movie_sorts = ["created_at.asc", "created_at.desc", "release_date.asc", "release_date.desc", "title.asc", "title.desc", "vote_average.asc", "vote_average.desc"]
v4_show_sorts = ["created_at.asc", "created_at.desc", "first_air_date.asc", "first_air_date.desc", "name.asc", "name.desc", "vote_average.asc", "vote_average.desc"]
v4_list_sorts = ["original_order.asc", "original_order.desc", "vote_average.asc", "vote_average.desc", "primary_release_date.asc", "primary_release_date.desc", "title.asc", "title.desc"]

class TMDbPagination(TMDbObj):
    """ Represents a Pagination Object. The standard iterator only loops through the current page.

        Attributes:
            page (int): Current Page.
            results (int): Current Page's Results.
            total_pages (int): Total Pages.
            total_results (int): Total Results.
    """
    def __init__(self, tmdb, data, value_type, page_type):
        self._value_type = value_type
        self._page_type = page_type
        self._results_text = None
        self._total_results_text = None
        self._page_storage = {}
        self._params = parse_qs(urlparse(tmdb._api.response.url).query)
        self._loading = True
        self.page = None
        self._loading = False
        super().__init__(tmdb=tmdb, data=data)

    def _load(self, data):
        self._partial = data is not None
        super()._load(self._get_page(1) if data is None else data)
        self.page = self._parse(attrs="page", value_type="int")
        self.total_pages = self._parse(attrs="total_pages", value_type="int")
        if self._total_results_text:
            self.total_results = self._parse(attrs=self._total_results_text, value_type="int")
            setattr(self, self._total_results_text, self.total_results)
        else:
            self.total_results = self._parse(attrs="total_results", value_type="int")
        if self._results_text:
            self.results = self._parse(attrs=self._results_text, value_type=self._value_type, is_list=True)
            setattr(self, self._results_text, self.results)
        else:
            self.results = self._parse(attrs="results", value_type=self._value_type, is_list=True)
        self._page_storage[self.page] = self.results
        self._finish(self._page_type)

    def load_next(self):
        """ Loads the next page of the Paginated Object.

            Returns:
                List[:class:`~tmdbapis.objs.pagination.Movie`]: List of Movies that changed.

            Raises:
                :class:`~tmdbapis.exceptions.Invalid`: When ``start_date`` or ``end_date`` is in an incorrect format.
        """
        if self.page + 1 > self.total_pages or self.page + 1 > 500:
            raise NotFound("No Next Page")
        self.load_page(self.page + 1)

    def load_page(self, page: int):
        """ Loads the page of the Paginated Object.

            Parameters:
                page (int): page number to load.

            Raises:
                :class:`~tmdbapis.exceptions.Invalid`: When ``page`` is not in the range of valid page numbers.
        """
        page = int(page)
        if page < 1 or page > self.total_pages or page > 500:
            raise Invalid(f"Page must be an integer 1-{'500' if self.total_pages > 500 else self.total_pages}")
        if page in self._page_storage:
            self._loading = True
            self.results = self._page_storage[page]
            self.page = page
            self._loading = False
        else:
            self._load(self._get_page(page))

    def __iter__(self):
        return (o for o in self.results)

    def __getitem__(self, index):
        return self.results[index]

    def __len__(self):
        return self.total_results

    @abstractmethod
    def _get_page(self, page):
        """ abstract get page """

    def get_results(self, amount: Optional[int] = None):
        """ Gets the amount of results asked for from multiple pages. This method can make alot of calls to the API if you're not careful.

            Parameters:
                amount (Optional[int]): Amount of Items you want returned.

            Raises:
                :class:`~tmdbapis.exceptions.Invalid`: When ``amount`` is not greater than zero.
         """
        if self.total_results == 0:
            return []
        if amount is None or int(amount) > self.total_results:
            amount = self.total_results
        if amount < 1:
            raise Invalid("amount must be greater than 0")
        results = []
        current_page = 0
        while len(results) < amount and current_page < self.total_pages and current_page < 500:
            current_page += 1
            self.load_page(current_page)
            if len(results) + len(self.results) < amount:
                results.extend(self.results)
            else:
                results.extend(self.results[:amount-len(results)])
        return results

    def _sort(self, sort_by, v3, is_movie):
        if not sort_by:
            return None
        elif v3 and sort_by not in v3_sorts:
            raise Invalid(f"sort_by not in {v3_sorts}")
        elif not v3 and is_movie and sort_by not in v4_movie_sorts:
            raise Invalid(f"sort_by not in {v4_movie_sorts}")
        elif not v3 and not is_movie and sort_by not in v4_show_sorts:
            raise Invalid(f"sort_by not in {v4_show_sorts}")
        else:
            return sort_by


class CreatedLists(TMDbPagination):
    """ Paginated Object of the lists created by an account. Will include private lists if you are the owner. """
    def __init__(self, tmdb, v3=False):
        self._v3 = v3 or tmdb._api4 is None
        super().__init__(tmdb, None, "list", "CreatedLists")

    def _get_page(self, page):
        if self._v3 or not self._api4.has_write_token:
            return self._api.account_get_created_lists(language=self._tmdb.language, page=page)
        else:
            return self._api4.account_get_lists(page=page)


class DiscoverMovies(TMDbPagination):
    """ Paginated Object of the Movie Discover search results. """
    def __init__(self, tmdb, **kwargs):
        self._kwargs = kwargs
        self._kwargs["language"] = tmdb.language
        super().__init__(tmdb, None, "movie", "DiscoverMovies")

    def _get_page(self, page):
        self._kwargs["page"] = page
        return self._api.discover_movie_discover(**self._kwargs)


class DiscoverTVShows(TMDbPagination):
    """ Paginated Object of the TV Discover search results. """
    def __init__(self, tmdb, **kwargs):
        self._kwargs = kwargs
        self._kwargs["language"] = tmdb.language
        super().__init__(tmdb, None, "tv", "DiscoverTVShows")

    def _get_page(self, page):
        self._kwargs["page"] = page
        return self._api.discover_tv_discover(**self._kwargs)


class FavoriteMovies(TMDbPagination):
    """ Paginated Object of your favorite movies.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None, v3=False):
        self._v3 = v3 or tmdb._api4 is None
        self._loading = True
        self.sort_by = self._sort(sort_by, self._v3, True)
        self._loading = False
        super().__init__(tmdb, None, "movie", "FavoriteMovies")

    def _get_page(self, page):
        if self._v3 or not self._api4.has_write_token:
            return self._api.account_get_favorite_movies(language=self._tmdb.language, sort_by=self.sort_by, page=page)
        else:
            return self._api4.account_get_favorite_movies(sort_by=self.sort_by, page=page)


class FavoriteTVShows(TMDbPagination):
    """ Paginated Object of your favorite TV shows.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None, v3=False):
        self._v3 = v3 or tmdb._api4 is None
        self._loading = True
        self.sort_by = self._sort(sort_by, self._v3, False)
        self._loading = False
        super().__init__(tmdb, None, "tv", "FavoriteTVShows")

    def _get_page(self, page):
        if self._v3 or not self._api4.has_write_token:
            return self._api.account_get_favorite_tv_shows(language=self._tmdb.language, sort_by=self.sort_by, page=page)
        else:
            return self._api4.account_get_favorite_tv_shows(sort_by=self.sort_by, page=page)


class RatedEpisodes(TMDbPagination):
    """ Paginated Object of the TV episodes you have rated.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None):
        self._loading = True
        self.sort_by = self._sort(sort_by, True, False)
        self._loading = False
        super().__init__(tmdb, None, "episode", "RatedEpisodes")

    def _get_page(self, page):
        return self._api.account_get_rated_tv_episodes(language=self._tmdb.language, sort_by=self.sort_by, page=page)


class RatedMovies(TMDbPagination):
    """ Paginated Object of the movies you have rated.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None, v3=False):
        self._v3 = v3 or tmdb._api4 is None
        self._loading = True
        self.sort_by = self._sort(sort_by, self._v3, True)
        self._loading = False
        super().__init__(tmdb, None, "movie", "RatedMovies")

    def _get_page(self, page):
        if self._v3 or not self._api4.has_write_token:
            return self._api.account_get_rated_movies(language=self._tmdb.language, sort_by=self.sort_by, page=page)
        else:
            return self._api4.account_get_rated_movies(sort_by=self.sort_by, page=page)


class RatedTVShows(TMDbPagination):
    """ Paginated Object of the TV shows you have rated.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None, v3=False):
        self._v3 = v3 or tmdb._api4 is None
        self._loading = True
        self.sort_by = self._sort(sort_by, self._v3, False)
        self._loading = False
        super().__init__(tmdb, None, "tv", "RatedTVShows")

    def _get_page(self, page):
        if self._v3 or not self._api4.has_write_token:
            return self._api.account_get_rated_tv_shows(language=self._tmdb.language, sort_by=self.sort_by, page=page)
        else:
            return self._api4.account_get_rated_tv_shows(sort_by=self.sort_by, page=page)


class MovieLists(TMDbPagination):
    """ Paginated Object of the lists that this movie belongs to.

        Attributes:
            movie_id (int): Movie ID referenced by this Paginated Object.
    """
    def __init__(self, tmdb, data, movie_id):
        self._loading = True
        self.movie_id = movie_id
        self._loading = False
        super().__init__(tmdb, data, "list", "MovieLists")

    def _get_page(self, page):
        return self._api.movies_get_lists(self.movie_id, language=self._tmdb.language, page=page)


class MovieRecommendations(TMDbPagination):
    """ Paginated Object of your personal movie recommendations.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None):
        self._loading = True
        self.sort_by = self._sort(sort_by, False, True)
        self._loading = False
        super().__init__(tmdb, None, "movie", "MovieRecommendations")

    def _get_page(self, page):
        return self._tmdb._v4_check(write=True).account_get_movie_recommendations(sort_by=self.sort_by, page=page)


class MovieReviews(TMDbPagination):
    """ Paginated Object of user reviews for a movie.

        Attributes:
            movie_id (int): Movie ID referenced by this Paginated Object.
    """
    def __init__(self, tmdb, data, movie_id):
        self._loading = True
        self.movie_id = movie_id
        self._loading = False
        super().__init__(tmdb, data, "review", "MovieReviews")

    def _get_page(self, page):
        return self._api.movies_get_reviews(self.movie_id, language=self._tmdb.language, page=page)


class MovieWatchlist(TMDbPagination):
    """ Paginated Object of the movies you have added to your watchlist.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None, v3=False):
        self._v3 = v3 or tmdb._api4 is None
        self._loading = True
        self.sort_by = self._sort(sort_by, self._v3, True)
        self._loading = False
        super().__init__(tmdb, None, "movie", "MovieWatchlist")

    def _get_page(self, page):
        if self._v3 or not self._api4.has_write_token:
            return self._api.account_get_movie_watchlist(language=self._tmdb.language, sort_by=self.sort_by, page=page)
        else:
            return self._api4.account_get_movie_watchlist(sort_by=self.sort_by, page=page)


class NowPlayingMovies(TMDbPagination):
    """ Paginated Object of movies in theatres.

        Attributes:
            region (str): ISO 3166-1 code used to filter release dates.
            maximum_date (datetime): Maximum Date.
            minimum_date (datetime): Minimum Date.
    """
    def __init__(self, tmdb, region=None):
        self._loading = True
        self.region = region
        self._loading = False
        super().__init__(tmdb, None, "movie", "NowPlayingMovies")

    def _load(self, data):
        super()._load(data)
        self._loading = True
        self.maximum_date = self._parse(attrs=["dates", "maximum"], value_type="date")
        self.minimum_date = self._parse(attrs=["dates", "minimum"], value_type="date")
        self._loading = False

    def _get_page(self, page):
        return self._api.movies_get_now_playing(language=self._tmdb.language, region=self.region, page=page)


class PopularMovies(TMDbPagination):
    """ Paginated Object of the current popular movies on TMDB.

        Attributes:
            region (str): ISO 3166-1 code used to filter release dates.
    """
    def __init__(self, tmdb, region=None):
        self._loading = True
        self.region = region
        self._loading = False
        super().__init__(tmdb, None, "movie", "PopularMovies")

    def _get_page(self, page):
        return self._api.movies_get_popular(language=self._tmdb.language, region=self.region, page=page)


class PopularPeople(TMDbPagination):
    """ Paginated Object of the list of popular people on TMDB. """
    def __init__(self, tmdb):
        super().__init__(tmdb, None, "person", "PopularPerson")

    def _get_page(self, page):
        return self._api.people_get_popular(self._tmdb.language, page=page)


class PopularTVShows(TMDbPagination):
    """ Paginated Object of the current popular TV shows on TMDB. """
    def __init__(self, tmdb):
        super().__init__(tmdb, None, "tv", "PopularTVShows")

    def _get_page(self, page):
        return self._api.tv_get_popular(language=self._tmdb.language, page=page)


class RecommendedMovies(TMDbPagination):
    """ Paginated Object of recommended movies for a movie.

        Attributes:
            movie_id (int): Movie ID referenced by this Paginated Object.
    """
    def __init__(self, tmdb, data, movie_id):
        self._loading = True
        self.movie_id = movie_id
        self._loading = False
        super().__init__(tmdb, data, "movie", "RecommendedMovies")

    def _get_page(self, page):
        return self._api.movies_get_recommendations(self.movie_id, language=self._tmdb.language, page=page)


class RecommendedTVShows(TMDbPagination):
    """ Paginated Object of recommended TV shows for a TV show.

        Attributes:
            tv_id (int): TV ID referenced by this Paginated Object.
    """
    def __init__(self, tmdb, data, tv_id):
        self._loading = True
        self.tv_id = tv_id
        self._loading = False
        super().__init__(tmdb, data, "tv", "RecommendedTVShows")

    def _get_page(self, page):
        return self._api.tv_get_recommendations(self.tv_id, language=self._tmdb.language, page=page)


class SearchCollections(TMDbPagination):
    """ Paginated Object of the Collections Search Results.

        Attributes:
            query (str): Query of the search.
    """
    def __init__(self, tmdb, query):
        self._loading = True
        self.query = query
        self._loading = False
        super().__init__(tmdb, None, "collection", "SearchCollections")

    def _get_page(self, page):
        results = self._api.search_search_collections(self.query, language=self._tmdb.language, page=page)
        if int(results["total_results"]) > 0:
            return results
        raise NotFound("No Results Found")


class SearchCompanies(TMDbPagination):
    """ Paginated Object of the Companies Search Results.

        Attributes:
            query (str): Query of the search.
    """
    def __init__(self, tmdb, query):
        self._loading = True
        self.query = query
        self._loading = False
        super().__init__(tmdb, None, "company", "SearchCompanies")

    def _get_page(self, page):
        results = self._api.search_search_companies(self.query, page=page)
        if int(results["total_results"]) > 0:
            return results
        raise NotFound("No Results Found")


class SearchKeywords(TMDbPagination):
    """ Paginated Object of the Keywords Search Results.

        Attributes:
            query (str): Query of the search.
    """
    def __init__(self, tmdb, query):
        self._loading = True
        self.query = query
        self._loading = False
        super().__init__(tmdb, None, "keyword", "SearchKeywords")

    def _get_page(self, page):
        results = self._api.search_search_keywords(self.query, page=page)
        if int(results["total_results"]) > 0:
            return results
        raise NotFound("No Results Found")


class SearchMovies(TMDbPagination):
    """ Paginated Object of the Movies Search Results.

        Attributes:
            query (str): Query of the search.
            include_adult (bool): Adult results in the search.
            region (str): ISO 3166-1 code used to filter the search.
            year (int): Year used to filter the search.
            primary_release_year (int): Primary Release Year used to filter the search.
    """
    def __init__(self, tmdb, query, include_adult=None, region=None, year=None, primary_release_year=None):
        self._loading = True
        self.query = query
        self.include_adult = include_adult
        self.region = region
        self.year = year
        self.primary_release_year = primary_release_year
        self._loading = False
        super().__init__(tmdb, None, "movie", "SearchMovies")

    def _get_page(self, page):
        results = self._api.search_search_movies(
            self.query,
            language=self._tmdb.language,
            page=page,
            include_adult=self.include_adult,
            region=self.region,
            year=self.year,
            primary_release_year=self.primary_release_year
        )
        if int(results["total_results"]) > 0:
            return results
        raise NotFound("No Results Found")


class SearchMulti(TMDbPagination):
    """ Paginated Object of the Multi Search Results.

        Attributes:
            query (str): Query of the search.
            include_adult (bool): Adult results in the search.
            region (str): ISO 3166-1 code used to filter the search.
    """
    def __init__(self, tmdb, query, include_adult=None, region=None):
        self._loading = True
        self.query = query
        self.include_adult = include_adult
        self.region = region
        self._loading = False
        super().__init__(tmdb, None, "media_type", "SearchMulti")

    def _get_page(self, page):
        results = self._api.search_multi_search(
            self.query,
            language=self._tmdb.language,
            page=page,
            include_adult=self.include_adult,
            region=self.region
        )
        if int(results["total_results"]) > 0:
            return results
        raise NotFound("No Results Found")


class SearchPeople(TMDbPagination):
    """ Paginated Object of the People Search Results.

        Attributes:
            query (str): Query of the search.
            include_adult (bool): Adult results in the search.
            region (str): ISO 3166-1 code used to filter the search.
    """
    def __init__(self, tmdb, query, include_adult=None, region=None):
        self._loading = True
        self.query = query
        self.include_adult = include_adult
        self.region = region
        self._loading = False
        super().__init__(tmdb, None, "person", "SearchPeople")

    def _get_page(self, page):
        results = self._api.search_search_people(
            self.query,
            language=self._tmdb.language,
            page=page,
            include_adult=self.include_adult,
            region=self.region
        )
        if int(results["total_results"]) > 0:
            return results
        raise NotFound("No Results Found")


class SearchTVShows(TMDbPagination):
    """ Paginated Object of the TV Show Search Results.

        Attributes:
            query (str): Query of the search.
            include_adult (bool): Adult results in the search.
            first_air_date_year (int): First Air Date Year used to filter the search.
    """
    def __init__(self, tmdb, query, include_adult=None, first_air_date_year=None):
        self._loading = True
        self.query = query
        self.include_adult = include_adult
        self.first_air_date_year = first_air_date_year
        self._loading = False
        super().__init__(tmdb, None, "tv", "SearchTVShows")

    def _get_page(self, page):
        results = self._api.search_search_tv_shows(
            self.query,
            language=self._tmdb.language,
            page=page,
            include_adult=self.include_adult,
            first_air_date_year=self.first_air_date_year
        )
        if int(results["total_results"]) > 0:
            return results
        raise NotFound("No Results Found")


class SimilarMovies(TMDbPagination):
    """ Paginated Object of similar movies for a movie.

        Attributes:
            movie_id (int): Movie ID referenced by this Paginated Object.
    """
    def __init__(self, tmdb, data, movie_id):
        self._loading = True
        self.movie_id = movie_id
        self._loading = False
        super().__init__(tmdb, data, "movie", "SimilarMovies")

    def _get_page(self, page):
        return self._api.movies_get_similar_movies(self.movie_id, language=self._tmdb.language, page=page)


class SimilarTVShows(TMDbPagination):
    """ Paginated Object of similar TV shows for a TV show.

        Attributes:
            tv_id (int): TV ID referenced by this Paginated Object.
    """
    def __init__(self, tmdb, data, tv_id):
        self._loading = True
        self.tv_id = tv_id
        self._loading = False
        super().__init__(tmdb, data, "tv", "SimilarTVShows")

    def _get_page(self, page):
        return self._api.tv_get_similar_tv_shows(self.tv_id, language=self._tmdb.language, page=page)


class TaggedImages(TMDbPagination):
    """ Paginated Object of tagged images for a person.

        Attributes:
            person_id (int): Person ID referenced by this Paginated Object.
    """
    def __init__(self, tmdb, data, person_id):
        self._loading = True
        self.person_id = person_id
        self._loading = False
        super().__init__(tmdb, data, "tagged", "TaggedImages")

    def _get_page(self, page):
        return self._api.people_get_tagged_images(self.person_id, language=self._tmdb.language, page=page)


class TMDbList(TMDbPagination):
    """ Paginated Object of the items in a list.

        Attributes:
            average_rating (float): Average Rating of the items in the List. (v4 Only)
            backdrop_path (str): List backdrop path. (v4 Only)
            backdrop_url (str): List full backdrop url. (v4 Only)
            comments (Dict[str, str]): Dictionary of the comment for each item. (v4 Only)
            country (:class:`~tmdbapis.objs.simple.Country`): Country object for the ISO 3166-1 Country Code. (v4 Only)
            created_by (Union[str, :class:`~tmdbapis.objs.simple.User`]): User who created the list (str for v3 and :class:`~tmdbapis.objs.simple.User` for v4).
            description (str): List Description.
            favorite_count (int): Number of users who have marked this list as a favorite. (v3 Only)
            id (int): List ID.
            iso_3166_1 (str): ISO 3166-1 Alpha-2 Country Code of the List. (v4 Only)
            iso_639_1 (str): ISO 639-1 Language Code of the List.
            language (:class:`~tmdbapis.objs.simple.Language`): Language object for the ISO 639-1 Language Code.
            name (int): List Name.
            object_ids (Dict[str, str]): Dictionary of the object ids for each item. (v4 Only)
            poster_path (str): List poster path.
            poster_url (str): List full poster url.
            public (bool): List Public. (v4 Only)
            revenue (int): Total revenue of the items in the List. (v4 Only)
            runtime (int): Total runtime of the items in the List. (v4 Only)
            sort_by (str): How the List is sorted. (v4 Only)
    """
    def __init__(self, tmdb, data, sort_by=None, load=False):
        self._sort_by = sort_by
        super().__init__(tmdb, data, "media_type", "List")
        if load:
            self._load(None)

    def _load(self, data):
        self._partial = data is not None
        self._results_text = None if self._api4 else "items"
        self._total_results_text = None if self._api4 else "item_count"
        super()._load(self._get_page(1) if data is None else data)
        self._loading = True
        self.description = self._parse(attrs="description")
        self.id = self._parse(attrs="id", value_type="int")
        self.iso_639_1 = self._parse(attrs="iso_639_1")
        self.language = self._tmdb._get_object(self._data, "language")
        self.name = self._parse(attrs="name")
        self.poster_path = self._parse(attrs="poster_path")
        self.poster_url = self._image_url(self.poster_path)
        if self._api4:
            self.average_rating = self._parse(attrs="average_rating", value_type="float")
            self.backdrop_path = self._parse(attrs="backdrop_path")
            self.backdrop_url = self._image_url(self.backdrop_path)
            self.comments = self._parse(attrs="comments", value_type="dict")
            self.country = self._tmdb._get_object(self._data, "country")
            self.created_by = self._parse(attrs="created_by", value_type="user")
            self.iso_3166_1 = self._parse(attrs="iso_3166_1")
            self.object_ids = self._parse(attrs="object_ids", value_type="dict")
            self.public = self._parse(attrs="public", value_type="bool")
            self.revenue = self._parse(attrs="revenue", value_type="int")
            self.runtime = self._parse(attrs="runtime", value_type="int")
            self.sort_by = self._parse(attrs="sort_by")
        else:
            self.created_by = self._parse(attrs="created_by")
            self.favorite_count = self._parse(attrs="favorite_count", value_type="int")
        self._loading = False

    def _get_page(self, page):
        if self._api4:
            return self._api4.list_get_list(
                self.id,
                language=self._tmdb.language,
                sort_by=self._sort_by,
                page=page
            )
        else:
            return self._api.lists_get_details(
                self.id,
                language=self._tmdb.language,
                page=page
            )

    def reload(self):
        """ Reloads the full object. """
        self._load(None)

    def update(self, name: Optional[str] = None, description: Optional[str] = None,
               public: Optional[bool] = None, sort_by: Optional[str] = None):
        """ Updates the list's metadata.

            Parameters:
                name (Optional[str]): Updates the list name.
                description (Optional[str]): Updates the list description.
                public (Optional[bool]): Updates whether the list is public.
                sort_by (Optional[str]): Updates the list sort_by.

            Raises:
                :class:`~tmdbapis.exceptions.Invalid`: When no attribute is given or the ``sort_by`` given is invalid.
        """
        if name is None and description is None and public is None and sort_by is None:
            raise Invalid("Must have at least one parameter to update (name, description, public, or sort_by)")
        if sort_by and sort_by not in v4_list_sorts:
            raise Invalid(f"sort_by not in {v4_list_sorts}")
        self._tmdb._v4_check().list_update_list(
            self.id,
            name=name,
            description=description,
            public=public,
            sort_by=sort_by
        )
        self.reload()

    def _check_item(self, item):
        if isinstance(item, Movie):
            item_id = item.id
            item_type = "movie"
        elif isinstance(item, TVShow):
            item_id = item.id
            item_type = "tv"
        elif isinstance(item, tuple):
            item_id = int(item[0])
            item_type = item[1]
            if item_type not in ["movie", "tv"]:
                raise Invalid("Tuple must have either 'movie' or 'tv'.")
        else:
            raise Invalid("Item must be either a Movie Object, a TV Show object, or a Tuple of the ID and either 'movie' or 'tv'.")
        if item_type == "tv":
            self._tmdb._v4_check()
        return item_id, item_type

    def has_item(self, item: Union[Movie, TVShow, Tuple[int, str]]) -> bool:
        """ Check to see if the Item given is in the List.

            Parameters:
                item (Union[Movie, TVShow, Tuple[int, str]]): Item you want to check if it's on the list. If not using a Movie or Show object then you must use a Tuple with the ID and either ``movie`` or ``tv``.

            Raises:
                :class:`~tmdbapis.exceptions.Invalid`: When the item given isn't in a valid format.

            Returns:
                bool: Whether the list has the Item or not.
        """
        item_id, item_type = self._check_item(item)
        if self._api4 and self._api4.has_write_token:
            try:
                self._api4.list_check_item_status(self.id, item_id, item_type)
                return True
            except NotFound:
                return False
        else:
            return self._api.lists_check_item_status(self.id, item_id)["item_present"]

    def add_items(self, items: List[Union[Movie, TVShow, Tuple[int, str]]]):
        """ Adds the items given to the list.

            Parameters:
                items (List[Union[Movie, TVShow, Tuple[int, str]]]): Items you want to add to the list. If not using a Movie or Show object then you must use a Tuple with the ID and either ``movie`` or ``tv``.

            Raises:
                :class:`~tmdbapis.exceptions.Invalid`: When the item given isn't in a valid format.

            Returns:
                bool: Whether the list has the Item or not.
        """
        item_ids = []
        if not isinstance(items, list):
            items = [items]
        for item in items:
            item_id, item_type = self._check_item(item)
            item_ids.append({"media_type": item_type, "media_id": item_id})

        if self._api4 and self._api4.has_write_token:
            self._api4.list_add_items(self.id, item_ids)
        else:
            for item_id in item_ids:
                self._api.lists_add_movie(self.id, item_id["media_id"])
        self.reload()

    def remove_items(self, items: List[Union[Movie, TVShow, Tuple[int, str]]]):
        """ Adds the items given to the list.

            Parameters:
                items (List[Union[Movie, TVShow, Tuple[int, str]]]): Item you want to remove from the list. If not using a Movie or Show object then you must use a Tuple with the ID and either ``movie`` or ``tv``.

            Raises:
                :class:`~tmdbapis.exceptions.Invalid`: When the item given isn't in a valid format.
        """
        item_ids = []
        if not isinstance(items, list):
            items = [items]
        for item in items:
            item_id, item_type = self._check_item(item)
            item_ids.append({"media_type": item_type, "media_id": item_id})

        if self._api4 and self._api4.has_write_token:
            self._api4.list_remove_items(self.id, item_ids)
        else:
            for item_id in item_ids:
                self._api.lists_remove_movie(self.id, item_id["media_id"])
        self.reload()

    def update_items(self, items: List[Tuple[Union[Movie, TVShow, Tuple[int, str]], str]]):
        """ Updates the items on the list.

            Parameters:
                items (List[Tuple[Union[Movie, TVShow, Tuple[int, str]], str]]): Tuples of the items you want updated on the list and their description. If not using a Movie or Show object then you must use a Tuple of the ID and either ``movie`` or ``tv``.

            Raises:
                :class:`~tmdbapis.exceptions.Invalid`: When the item given isn't in a valid format.
        """
        item_ids = []
        if not isinstance(items, list):
            items = [items]
        for item in items:
            item_id, item_type = self._check_item(item[0])
            comment = item[1]
            item_ids.append({"media_type": item_type, "media_id": item_id, "comment": comment})
        self._tmdb._v4_check(write=True).list_update_items(self.id, item_ids)
        self.reload()

    def clear(self):
        """ Clear all items from the list. """
        if self._api4 and self._api4.has_write_token:
            self._api4.list_clear_list(self.id)
        else:
            self._api.lists_clear_list(self.id, confirm=True)
        self.reload()

    def delete(self):
        """ Delete the list. """
        if self._api4 and self._api4.has_write_token:
            self._api4.list_delete_list(self.id)
        else:
            try:
                self._api.lists_delete_list(self.id)
            except TMDbException as e:
                if not str(e).startswith("(500 [Internal Server Error])"):
                    raise


class TopRatedMovies(TMDbPagination):
    """ Paginated Object of the top rated movies on TMDB.

        Attributes:
            region (str): ISO 3166-1 code used to filter release dates.
    """
    def __init__(self, tmdb, region=None):
        self._loading = True
        self.region = region
        self._loading = False
        super().__init__(tmdb, None, "movie", "TopRatedMovies")

    def _get_page(self, page):
        return self._api.movies_get_top_rated(language=self._tmdb.language, region=self.region, page=page)


class TopRatedTVShows(TMDbPagination):
    """ Paginated Object of the top rated TV shows on TMDB. """
    def __init__(self, tmdb):
        super().__init__(tmdb, None, "tv", "TopRatedTVShows")

    def _get_page(self, page):
        return self._api.tv_get_top_rated(language=self._tmdb.language, page=page)


class Trending(TMDbPagination):
    """ Paginated Object of daily or weekly trending items.

        Attributes:
            media_type (str): Media type of the Paginated Object.
            time_window (str): Time Window for the Trending.
    """
    def __init__(self, tmdb, media_type, time_window):
        self._loading = True
        self.media_type = media_type
        self.time_window = time_window
        self._loading = False
        super().__init__(tmdb, None, "media_type", "Trending")

    def _get_page(self, page):
        return self._api.trending_get_trending(self.media_type, self.time_window, page=page)


class TVShowRecommendations(TMDbPagination):
    """ Paginated Object of your personal TV show recommendations.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None):
        self._loading = True
        self.sort_by = self._sort(sort_by, False, False)
        self._loading = False
        super().__init__(tmdb, None, "tv", "TVShowRecommendations")

    def _get_page(self, page):
        return self._tmdb._v4_check(write=True).account_get_tv_show_recommendations(sort_by=self.sort_by, page=page)


class TVShowsAiringToday(TMDbPagination):
    """ Paginated Object of TV shows that are airing today. """
    def __init__(self, tmdb):
        super().__init__(tmdb, None, "tv", "TopRatedMovies")

    def _get_page(self, page):
        return self._api.tv_get_tv_airing_today(language=self._tmdb.language, page=page)


class TVShowsOnTheAir(TMDbPagination):
    """ Paginated Object of shows that are currently on the air. """
    def __init__(self, tmdb):
        super().__init__(tmdb, None, "tv", "TVShowsOnTheAir")

    def _get_page(self, page):
        return self._api.tv_get_tv_on_the_air(language=self._tmdb.language, page=page)


class TVShowWatchlist(TMDbPagination):
    """ Paginated Object of the TV shows you have added to your watchlist.

        Attributes:
            sort_by (str): How the paginated object is sorted.
    """
    def __init__(self, tmdb, sort_by=None, v3=False):
        self._v3 = v3 or tmdb._api4 is None
        self._loading = True
        self.sort_by = self._sort(sort_by, self._v3, False)
        self._loading = False
        super().__init__(tmdb, None, "tv", "TVShowWatchlist")

    def _get_page(self, page):
        if self._v3 or not self._api4.has_write_token:
            return self._api.account_get_tv_show_watchlist(language=self._tmdb.language, sort_by=self.sort_by, page=page)
        else:
            return self._api4.account_get_tv_show_watchlist(sort_by=self.sort_by, page=page)


class UpcomingMovies(TMDbPagination):
    """ Paginated Object of upcoming movies in theatres.

        Attributes:
            maximum_date (datetime): Maximum Date.
            minimum_date (datetime): Minimum Date.
    """
    def __init__(self, tmdb, region=None):
        self._loading = True
        self.region = region
        self._loading = False
        super().__init__(tmdb, None, "movie", "UpcomingMovies")

    def _load(self, data):
        super()._load(data)
        self._loading = True
        self.maximum_date = self._parse(attrs=["dates", "maximum"], value_type="date")
        self.minimum_date = self._parse(attrs=["dates", "minimum"], value_type="date")
        self._loading = False

    def _get_page(self, page):
        return self._api.movies_get_upcoming(language=self._tmdb.language, region=self.region, page=page)



import logging
from json.decoder import JSONDecodeError
from typing import Dict, Optional, Union, List
from requests import Session, Response
from requests.exceptions import RequestException
from tmdbapis.exceptions import TMDbException, NotFound, Unauthorized, WritePermission, PrivateResource, \
    Authentication, Invalid

logger = logging.getLogger(__name__)

base_url = "https://api.themoviedb.org/4"

class API4:
    """ Raw V4 API Class containing all `TMDb API4 calls <https://developers.themoviedb.org/4/getting-started/authorization>`__.

        Parameters:
            access_token (str): TMDb V4 Access Token.
            session (Optional[Session]): :class:`requests.Session` object.
            validate (bool): Validate the TMDb V4 Access Token on creation.

        Attributes:
            access_token (str): TMDb V4 Access Token.
            has_write_token (bool): Does the provided TMDb V4 Access Token have write access.
            account_id (str): TMDb V4 Account ID.
            response (Response): TMDb V4 most recent full :class:`requests.Response` object.
    """
    def __init__(self, access_token: str, session: Optional[Session] = None, validate: bool = True):
        self.access_token = access_token
        self._account_id = None
        self._session = Session() if session is None else session
        self.response = None
        if validate:
            try:
                self.auth_create_access_token(self.access_token)
            except TMDbException:
                self.auth_create_request_token()

    def _get(self, path, **kwargs):
        """ process get request. """
        return self._request("get", path, **kwargs)

    def _delete(self, path, json=None, **kwargs):
        """ process delete request. """
        return self._request("delete", path, json=json, **kwargs)

    def _post(self, path, json=None, **kwargs):
        """ process post request. """
        return self._request("post", path, json=json, **kwargs)

    def _put(self, path, json=None, **kwargs):
        """ process put request. """
        return self._request("put", path, json=json, **kwargs)

    def _request(self, request_type, path, json=None, **kwargs):
        """ process request. """
        url_params = {k: v for k, v in kwargs.items() if v is not None}
        body_json = {k: v for k, v in json.items() if v is not None} if json else None
        request_url = f"{base_url}{path}"
        logger.debug(f"Request URL: {request_url}")
        headers = {"Authorization": f"Bearer {self.access_token}"}
        if body_json is not None:
            logger.debug(f"Request JSON: {body_json}")
        logger.debug(f"Headers: {headers}")
        try:
            if request_type == "delete":
                self.response = self._session.delete(request_url, json=body_json, headers=headers, params=url_params)
            elif request_type == "post":
                self.response = self._session.post(request_url, json=body_json, headers=headers, params=url_params)
            elif request_type == "put":
                self.response = self._session.put(request_url, json=body_json, headers=headers, params=url_params)
            else:
                self.response = self._session.get(request_url, headers=headers, params=url_params)
            response_json = self.response.json()
        except (RequestException, JSONDecodeError):
            raise TMDbException(f"Failed to Connect to {base_url}")
        logger.debug(f"Response ({self.response.status_code} [{self.response.reason}]) {response_json}")
        if self.response.status_code == 401:
            if "status_code" in response_json:
                if response_json["status_code"] == 36:
                    raise WritePermission("Requires V4 Authentication, use tmdbapis.v4_authenticate(), then approve the returned URL, and finally run tmdbapis.v4_approved()")
                elif response_json["status_code"] == 39:
                    raise PrivateResource(response_json["status_message"])
                elif response_json["status_code"] == 7:
                    raise Unauthorized(response_json["status_message"])
                else:
                    raise TMDbException(f"({response_json['status_code']}) {response_json['status_message']}")
            else:
                raise TMDbException(f"({self.response.status_code} [{self.response.reason}]) {response_json}")
        elif self.response.status_code == 404:
            raise NotFound(f"({self.response.status_code} [{self.response.reason}]) Requested Item Not Found")
        elif self.response.status_code >= 400:
            raise TMDbException(f"({self.response.status_code} [{self.response.reason}]) {response_json}")
        elif "errors" in response_json:
            raise TMDbException(response_json["errors"])
        elif "success" in response_json and response_json["success"] is False:
            raise TMDbException(response_json["status_message"])
        return response_json

    @property
    def has_write_token(self):
        return self._account_id is not None

    @property
    def account_id(self):
        if not self._account_id:
            raise Authentication(f"Requires V4 API Write Access Token, use tmdbapis.v4_access_token(access_token)")
        return self._account_id

    def account_get_lists(self, account_id: Optional[str] = None, page: Optional[int] = None) -> Dict:
        """ `Account Get Lists <https://developers.themoviedb.org/4/account/get-account-lists>`__

            Get all of the lists you've created.

            Parameters:
                account_id (Optional[str]): Account ID.
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/lists", page=page)

    def account_get_favorite_movies(
            self, account_id: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `Account Get Favorite Movies <https://developers.themoviedb.org/3/account/get-favorite-movies>`__

            Get the list of movies you have marked as a favorite.

            Parameters:
                account_id (Optional[str]): Account ID.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``created_at.asc``, ``created_at.desc``, ``release_date.asc``, ``release_date.desc``, ``title.asc``, ``title.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/movie/favorites", sort_by=sort_by, page=page)

    def account_get_favorite_tv_shows(
            self, account_id: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `Account Get Favorite TV Shows <https://developers.themoviedb.org/4/account/get-account-favorite-tv-shows>`__

            Get the list of TV shows you have marked as a favorite.

            Parameters:
                account_id (Optional[str]): Account ID.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``first_air_date.asc``, ``first_air_date.desc``, ``name.asc``, ``name.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/tv/favorites", sort_by=sort_by, page=page)

    def account_get_movie_recommendations(
            self, account_id: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `Account Get Movie Recommendations <https://developers.themoviedb.org/4/account/get-account-movie-recommendations>`__

            Get a list of your personal movie recommendations.

            Parameters:
                account_id (Optional[str]): Account ID.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``created_at.asc``, ``created_at.desc``, ``release_date.asc``, ``release_date.desc``, ``title.asc``, ``title.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/movie/recommendations", sort_by=sort_by, page=page)

    def account_get_tv_show_recommendations(
            self, account_id: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `Account Get TV Show Recommendations <https://developers.themoviedb.org/4/account/get-account-tv-show-recommendations>`__

            Get a list of your personal TV show recommendations.

            Parameters:
                account_id (Optional[str]): Account ID.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``first_air_date.asc``, ``first_air_date.desc``, ``name.asc``, ``name.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/tv/rated", sort_by=sort_by, page=page)

    def account_get_movie_watchlist(
            self, account_id: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `Account Get Movie Watchlist <https://developers.themoviedb.org/4/account/get-account-movie-watchlist>`__

            Get the list of movies you have added to your watchlist.

            Parameters:
                account_id (Optional[str]): Account ID.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``created_at.asc``, ``created_at.desc``, ``release_date.asc``, ``release_date.desc``, ``title.asc``, ``title.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/movie/watchlist", sort_by=sort_by, page=page)

    def account_get_tv_show_watchlist(
            self, account_id: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `Account Get TV Show Watchlist <https://developers.themoviedb.org/4/account/get-account-tv-show-watchlist>`__

            Get the list of TV shows you have added to your watchlist.

            Parameters:
                account_id (Optional[str]): Account ID.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``first_air_date.asc``, ``first_air_date.desc``, ``name.asc``, ``name.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/tv/watchlist", sort_by=sort_by, page=page)

    def account_get_rated_movies(
            self, account_id: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `Account Get Rated Movies <https://developers.themoviedb.org/4/account/get-account-rated-movies>`__

            Get the list of movies you have rated.

            Parameters:
                account_id (Optional[str]): Account ID.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``created_at.asc``, ``created_at.desc``, ``release_date.asc``, ``release_date.desc``, ``title.asc``, ``title.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/movie/rated", sort_by=sort_by, page=page)

    def account_get_rated_tv_shows(
            self, account_id: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `Account Get Rated TV Shows <https://developers.themoviedb.org/4/account/get-account-rated-tv-shows>`__

            Get the list of TV shows you have rated.

            Parameters:
                account_id (Optional[str]): Account ID.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``first_air_date.asc``, ``first_air_date.desc``, ``name.asc``, ``name.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/account/{account_id if account_id else self.account_id}/tv/rated", sort_by=sort_by, page=page)

    def auth_create_request_token(self, redirect_to: Optional[str] = None) -> Dict:
        """ `Auth Create Request Token <https://developers.themoviedb.org/4/auth/create-request-token>`__

            This method generates a new request token that you can ask a user to approve. This is the first step in
            getting permission from a user to read and write data on their behalf. You can read more about this system
            `here <https://developers.themoviedb.org/4/auth/user-authorization-1>`__.

            Parameters:
                redirect_to (Optional[str]): Redirect URL or callback that will be executed once a request token has been approved on TMDb.
        """
        return self._post("/auth/request_token", json={"redirect_to": redirect_to} if redirect_to else None)

    def auth_create_access_token(self, request_token: str) -> Dict:
        """ `Auth Create Access Token <https://developers.themoviedb.org/4/auth/create-access-token>`__

            This method will finish the user authentication flow and issue an official user access token. The request
            token in this request is sent along as part of the POST body. You should still use your standard API read
            access token for authenticating this request.

            Parameters:
                request_token (str): Request Token
        """
        response = self._post("/auth/access_token", json={"request_token": request_token})
        self.access_token = response["access_token"]
        self._account_id = response["account_id"]
        return response

    def auth_delete_access_token(self, access_token: str) -> Dict:
        """ `Auth Delete Access Token <https://developers.themoviedb.org/4/auth/delete-access-token>`__

            This method gives your users the ability to log out of a session.

            Parameters:
                access_token (str): Access Token
        """
        return self._delete("/auth/access_token", json={"access_token": access_token})

    def list_get_list(
            self, list_id: int,
            language: Optional[str] = None,
            sort_by: Optional[str] = None,
            page: Optional[str] = None
    ) -> Dict:
        """ `List Get List <https://developers.themoviedb.org/4/list/get-list>`__

            This method will retrieve a list by id.

            Private lists can only be accessed by their owners and therefore require a valid user access token.

            Parameters:
                list_id (int): List ID
                language (Optional[str]): ISO-639-1 or ISO-3166-1 value to display translated data for the fields that support it.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``original_order.asc``, ``original_order.desc``, ``release_date.asc``, ``release_date.desc``, ``title.asc``, ``title.desc``, ``vote_average.asc``, ``vote_average.desc``
                page (Optional[int]): Specify which page to query.
        """
        return self._get(f"/list/{list_id}", language=language, sort_by=sort_by, page=page)

    def list_create_list(
            self, name: str,
            iso_639_1: str,
            description: Optional[str] = None,
            public: Optional[bool] = None,
            iso_3166_1: Optional[str] = None
    ) -> Dict:
        """ `List Create List <https://developers.themoviedb.org/4/list/create-list>`__

            This method will create a new list.

            You will need to have valid user access token in order to create a new list.

            Parameters:
                name (str): Set the name of your list.
                iso_639_1 (str): Set the ISO-639-1 variant for your list.
                description (Optional[str]): Set the description of your list.
                public (Optional[bool]): Toggle the public status of your list.
                iso_3166_1 (Optional[int]): Set the ISO-3166-1 variant for your list.
        """
        return self._post(
            "/list",
            json={
                "name": name,
                "iso_639_1": iso_639_1,
                "description": description,
                "public": public,
                "iso_3166_1": iso_3166_1
            }
        )

    def list_update_list(
            self, list_id: int,
            name: Optional[str] = None,
            description: Optional[str] = None,
            public: Optional[bool] = None,
            sort_by: Optional[str] = None
    ) -> Dict:
        """ `List Update List <https://developers.themoviedb.org/4/list/update-list>`__

            This method will let you update the details of a list.

            You must be the owner of the list and therefore have a valid user access token in order to edit it.

            Parameters:
                list_id (int): List ID
                name (Optional[str]): Set the name of your list.
                description (Optional[str]): Set the description of your list.
                public (Optional[bool]): Toggle the public status of your list.
                sort_by (Optional[str]): Choose a sort option for the list of results. Allowed Values: ``original_order.asc``, ``original_order.desc``, ``vote_average.asc``, ``vote_average.desc``, ``primary_release_date.asc``, ``primary_release_date.desc``, ``title.asc``, ``title.desc``
        """
        return self._put(
            f"/list/{list_id}",
            json={
                "name": name,
                "description": description,
                "public": public,
                "sort_by": sort_by
            }
        )

    def list_clear_list(self, list_id: int) -> Dict:
        """ `List Clear List <https://developers.themoviedb.org/4/list/clear-list>`__

            This method lets you clear all of the items from a list in a single request. This action cannot be reversed so use it with caution.

            You must be the owner of the list and therefore have a valid user access token in order to clear a list.

            Parameters:
                list_id (int): List ID
        """
        return self._get(f"/list/{list_id}/clear")

    def list_delete_list(self, list_id: int) -> Dict:
        """ `List Clear List <https://developers.themoviedb.org/4/list/delete-list>`__

            This method will delete a list by id. This action is not reversible so take care when issuing it.

            You must be the owner of the list and therefore have a valid user access token in order to delete it.

            Parameters:
                list_id (int): List ID
        """
        return self._delete(f"/list/{list_id}")

    def _items(self, items, comment=False):
        json = {"items": []}
        for item in items:
            if "media_type" not in item or "media_id" not in item or item["media_type"] not in ["movie", "tv"] or \
                    (comment and "comment" not in item):
                error = "'media_type', 'media_id', and 'comment'" if comment else "'media_type' and 'media_id'"
                raise Invalid(f"Each object must have {error}. 'media_type' must be either 'movie' or 'tv'")
            json["items"].append({"media_type": item["media_type"], "media_id": int(item["media_id"])})
        return json

    def list_add_items(self, list_id: int, items: List[Dict[str, Union[str, int]]]) -> Dict:
        """ `List Add Items <https://developers.themoviedb.org/4/list/add-items>`__

            This method will let you add items to a list. We support essentially an unlimited number of items to be
            posted at a time. Both movie and TV series are support.

            The results of this query will return a ``results`` array. Each result includes a ``success`` field. If a
            result is ``false`` this will usually indicate that the item already exists on the list. It may also
            indicate that the item could not be found.

            You must be the owner of the list and therefore have a valid user access token in order to add items to a
            list.

            Parameters:
                list_id (int): List ID
                items (List[Dict[str, Union[str, int]]]): List of items to add. Each item is a dictionary with the format {"media_type": str, "media_id": int}. ``media_type`` can either be ``movie`` or ``tv``.
        """
        return self._post(f"/list/{list_id}/items", json=self._items(items))

    def list_update_items(self, list_id: int, items: List[Dict[str, Union[str, int]]]) -> Dict:
        """ `List Update Items <https://developers.themoviedb.org/4/list/update-items>`__

            This method will let you update an individual item on a list. Currently, only adding a comment is supported.

            You must be the owner of the list and therefore have a valid user access token in order to edit items.

            Parameters:
                list_id (int): List ID
                items (List[Dict[str, Union[str, int]]]): List of items to update. Each item is a dictionary with the format {"media_type": str, "media_id": int, "comment": str}. ``media_type`` can either be ``movie`` or ``tv``.
        """
        return self._put(f"/list/{list_id}/items", json=self._items(items, comment=True))

    def list_remove_items(self, list_id: int, items: List[Dict[str, Union[str, int]]]) -> Dict:
        """ `List Remove Items <https://developers.themoviedb.org/4/list/remove-items>`__

            This method will let you remove items from a list. You can remove multiple items at a time.

            You must be the owner of the list and therefore have a valid user access token in order to delete items from it.

            Parameters:
                list_id (int): List ID
                items (List[Dict[str, Union[str, int]]]): List of items to remove. Each item is a dictionary with the format {"media_type": str, "media_id": int}. ``media_type`` can either be ``movie`` or ``tv``.
        """
        return self._delete(f"/list/{list_id}/items", json=self._items(items))

    def list_check_item_status(self, list_id: int, media_id: int, media_type: str) -> Dict:
        """ `List Check Item Status <https://developers.themoviedb.org/4/list/check-item-status>`__

            This method lets you quickly check if the item is already added to the list.

            You must be the owner of the list and therefore have a valid user access token in order to check an item status.

            Parameters:
                list_id (int): List ID
                media_id (int): Media ID
                media_type (str): Set the kind of media object are you checking. Allowed Values: ``movie``, ``tv``
        """
        return self._get(f"/list/{list_id}/item_status", media_id=media_id, media_type=media_type)



import os, tmdbapis

from setuptools import setup, find_packages

with open("README.rst", "r") as f:
    long_descr = f.read()

__version__ = None
if os.path.exists("VERSION"):
    with open("VERSION") as handle:
        for line in handle.readlines():
            line = line.strip()
            if len(line) > 0:
                __version__ = line
                break

setup(
    name=tmdbapis.__package_name__,
    version=__version__,
    description=tmdbapis.__description__,
    long_description=long_descr,
    url=tmdbapis.__url__,
    author=tmdbapis.__author__,
    author_email=tmdbapis.__email__,
    license=tmdbapis.__license__,
    packages=find_packages(),
    python_requires=">=3.8",
    keywords=["tmdbapis", "tmdbapi", "tmdb", "wrapper", "api"],
    install_requires=[
      "requests"
    ],
    project_urls={
      "Documentation": "https://tmdbapis.kometa.wiki",
      "Funding": "https://github.com/sponsors/meisnate12",
      "Source": "https://github.com/Kometa-Team/TMDbAPIs",
      "Issues": "https://github.com/Kometa-Team/TMDbAPIs/issues",
    },
    classifiers=[
      "Development Status :: 5 - Production/Stable",
      "Intended Audience :: Developers",
      "Topic :: Software Development :: Libraries",
      "Programming Language :: Python",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
)

